
/******************************************************************************
* MODULE     : ext_convert.cc
* DESCRIPTION: Conversion routines between Mathemagix and C++
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
namespace mmx {

/******************************************************************************
* Compute the C/C++ name of a Mathemagix type. The table t contains
* already computed conversions for a finite number of additional types,
* usually template arguments
******************************************************************************/

generic
extender::type_as_ccpp (generic g, table t) {
  // search the C/C++ name of a Mathemagix type
  // some additional conversion rules may be specified in the table
  if (t -> contains (g)) return t[g];
  else if (type_conv -> contains (g)) return type_conv [g];
  else if (is_func (g, MMX_CROSS, 0)) return MMX_VOID;
  else if (is_list (g)) {
    list l= as_list (g), r;
    while (!nil (l)) {
      r= cons (type_as_ccpp (car (l), t), r);
      l= cdr (l);
    }
    return reverse (r);
  }
  else if (g == MMX_OSTREAM) return Alias (g);
  else if ((g == MMX_TUPLETYPE) || (g == MMX_VTUPLETYPE))
    return Scope (symbol ("mmx"), symbol ("glued_tuple"));
  else if ((g == MMX_CLASS) ||
	   (g == MMX_VOID) ||
	   (g == MMX_BOOL) ||
	   (g == MMX_INT) ||
	   (g == MMX_LITERAL_INT) ||
	   (g == MMX_LITERAL_FLOAT) ||
	   (g == MMX_LITERAL_STRING) ||
	   (g == MMX_ALIAS) ||
	   (g == MMX_CROSS) ||
	   (g == MMX_VCROSS) ||
	   (g == MMX_FUNCTION) ||
	   (g == CCPP_CONST) ||
	   (g == CCPP_VIRTUAL) ||
	   (g == geninst ()))
    return g;
  else {
    int i;
    for (i=1; i<=max_type_args; i++)
      if (g == geninst (i))
	return g;

    warning (g, "Type could not be translated");
    return g;
  }
}

list
extender::type_as_ccpp (list l, table t) {
  if (nil (l)) return l;
  return cons (type_as_ccpp (car (l), t), type_as_ccpp (cdr (l), t));
}

/******************************************************************************
* Convert a Mathemagix declaration into a C++ declaration. The gen_flag
* tells whether the declaration is a method of a generic type structure.
******************************************************************************/

generic
extender::decl_as_ccpp (generic g, bool gen_flag) {
  if (!is_func (g, MMX_DEFINE, 2))
    error (g, "Expected a declaration");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1][1]))
    error (g[1][1], "Expected a symbol");
  else if (!is_symbol (g[2]))
    error (g[2], "Expected a symbol");
  else if (!gen_flag)
    return list (g[1][0], g[2], type_as_ccpp (g[1][2]));
  else return list (g[1][0], genmethod (g[2]), type_as_ccpp (g[1][2]));
  return g;
}

list
extender::decl_as_ccpp (list l, bool gen_flag) {
  if (nil (l)) return l;
  else return cons (decl_as_ccpp (car (l), gen_flag),
		    decl_as_ccpp (cdr (l), gen_flag));
}

/******************************************************************************
* Viewing a Mathemagix generic as a C++ type
******************************************************************************/

bool
extender::is_geninst (generic g) {
  int i;
  for (i=1; i<=max_type_args; i++)
    if (g == geninst (i))
      return true;
  return g == geninst ();
}

generic
extender::view_as_ccpp (generic g, generic mmx_type, table t) {
  generic type= type_as_ccpp (mmx_type, t);
  generic cv  = mmx_specialize ("view_as", type);
  generic expr= g;
  if (mmx_type == MMX_CLASS) cv= symbol ("as_" * the_prefix * "_type");
  if (mmx_type == MMX_BOOL) cv= mmx_symbol ("as_bool");
  if (mmx_type == MMX_INT) cv= mmx_symbol ("as_int");
  if (mmx_type == MMX_OSTREAM) cv= mmx_symbol ("as_ostream");
  if (mmx_type == MMX_LITERAL_INT) cv= mmx_symbol ("as_literal_contents");
  if (mmx_type == MMX_LITERAL_FLOAT) cv= mmx_symbol ("as_literal_contents");
  if (mmx_type == MMX_LITERAL_STRING) cv= mmx_symbol ("as_literal_contents");
  if (is_func (mmx_type, MMX_ALIAS, 1))
    cv= mmx_specialize ("view_as_alias", type[1]);
  if (is_func (mmx_type, MMX_TUPLETYPE, 1)) {
    cv= mmx_specialize ("glued_tuple", type[1]);
    if (is_func (expr, MMX_ACCESS, 2)) // for security
      expr= expr[1]; // temporary hack to get arguments right
  }
  if (is_func (mmx_type, MMX_VTUPLETYPE, 1))
    cv= mmx_specialize ("glued_tuple", type[1]);
  if (is_geninst (type))
    cv= mmx_specialize ("view_as_inst", type);
  if (is_func (mmx_type, MMX_ALIAS, 1) && is_geninst (type[1]))
    cv= mmx_specialize ("view_as_alias_inst", type[1]);
  return list (cv, expr);
}

/******************************************************************************
* Viewing a C++ type as a Mathemagix generic
******************************************************************************/

generic
extender::view_as_mmx (generic expr, generic mmx_type, table t) {
  generic type= type_as_ccpp (mmx_type, t);
  generic cv  = mmx_specialize ("glued", type);
  if (mmx_type == MMX_CLASS) return Arrow (expr, symbol ("mmxt"));
  if (mmx_type == MMX_BOOL) cv= mmx_symbol ("single");
  if (mmx_type == MMX_INT) cv= mmx_symbol ("single");
  if (mmx_type == MMX_OSTREAM) cv= mmx_symbol ("printer");
  if (mmx_type == MMX_LITERAL_INT) cv= mmx_symbol ("literal_int");
  if (mmx_type == MMX_LITERAL_FLOAT) cv= mmx_symbol ("literal_float");
  if (mmx_type == MMX_LITERAL_STRING) cv= mmx_symbol ("literal_string");
  if (is_func (mmx_type, MMX_ALIAS, 1))
    cv= mmx_specialize ("glued_alias", type[1]);
  if (is_func (mmx_type, MMX_TUPLETYPE, 1))
    cv= mmx_specialize ("as_generic", type[1]);
  if (is_func (mmx_type, MMX_VTUPLETYPE, 1))
    cv= mmx_specialize ("as_generic", type[1]);
  return list (cv, expr);
}

list
extender::view_as_mmx (list exprs, list mmx_types, table t) {
  if (nil (exprs) || nil (mmx_types)) return list ();
  return cons (view_as_mmx (car (exprs), car (mmx_types), t),
	       view_as_mmx (cdr (exprs), cdr (mmx_types), t));
}

} // namespace mmx
