
/******************************************************************************
* MODULE     : ext_generic.cc
* DESCRIPTION: Produce the generic extension code
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
namespace mmx {

/******************************************************************************
* Specializing list of imports to a given type
******************************************************************************/

generic
extender::specialize (generic g, generic to, bool virtual_flag) {
  // Importations should all be of the form
  //   forall(T:Class) f: (A_1,...,A_n)->B == "f";
  // Genericity over more than one type is not yet implemented.
  // If one among A_1,...,A_n,B is of the form F(T),
  // where F(T) involves T, but is not equal to T or Alias T,
  // then a converter F(X)->F(Y) should be supplied for all X and Y
  // for which a converter X->Y exists.

  if (!is_func (g, MMX_FORALL, 2))
    error (g, "Expected a 'forall' type");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g[1], "Expected a typed quantification variable");
  else if (g[1][2] != MMX_CLASS)
    error (g[1][2], "Expected 'Class' type");
  else if (is_func (g[2], MMX_FORALL, 2))
    error (g, "Double quantifications not yet allowed");
  else if (!is_func (g[2], MMX_DEFINE, 2))
    error (g[2], "Expected a declaration");
  else if (!is_func (g[2][1], MMX_TYPE, 2))
    error (g[2], "Expected a declaration");
  else {
    generic qvar= g[1][1];
    generic body= g[2];
    generic ret = substitute (body, qvar, to);
    if (virtual_flag)
      ret= list (ret[0],
		 list (ret[1][0], ret[1][1], Virtual (ret[1][2])),
		 ret[2]);
    return ret;
  }
  return g;
}

list
extender::specialize (list l, generic to, bool virtual_flag) {
  if (nil (l)) return l;
  else return cons (specialize (car (l), to, virtual_flag),
		    specialize (cdr (l), to, virtual_flag));
}

/******************************************************************************
* Produce the gentype type. This type serves as a base type for
* all dynamic types which have been exported by the user.
******************************************************************************/

static list
arg_list (int start, int end) {
  if (start >= end) return list ();
  else return cons (parameter (start), arg_list (start+1, end));
}

list
extender::gentype_args_as_mmx (list l) {
  if (nil (l)) return l;
  generic var = car (l) [1];
  generic type= car (l) [2];
  generic rval;
  if (type == geninst ()) {
    generic ptr= Dot (var, symbol ("rep"));
    rval= Unpointer (Transtype (ptr, Pointer (mmx_generic)));
  }
  else rval = view_as_mmx (var, type, table ());
  generic rtype= build_mmx_type (type, list ());
  return cons (list (mmx_symbol ("value"), rval, rtype),
	       gentype_args_as_mmx (cdr (l)));
}

void
extender::gentype_methods (generic h) {
  generic g= decl_as_ccpp (h);
  if (!is_func (g, MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1]))
    error (g[1], "Expected a symbol");
  else if (!is_func (g[2], MMX_FUNCTION, 2))
    error (g[2], "Expected a function type");
  else if (!is_func (g[2][1], MMX_CROSS))
    error (g[2][1], "Expected a cross type");
  else {
    g= name_parameters (g);
    generic name= genmethod (g[1]);
    generic scoped= Scope (gentype_rep, name);
    generic fun= list (g[0], scoped, g[2]);
    if (glue_flag) {
      h= name_parameters (h[1]);
      list args= gentype_args_as_mmx (cdr (as_list (h[2][1])));
      string  sname= as_string (S (as_symbol (h[1])));
      generic symb = list (mmx_symbol ("symbol"), sname);
      generic instr= mmx_make_list (cons (symb, args));
      generic eval = list (mmx_symbol ("extern_eval"), instr);
      generic ret  = view_as_ccpp (eval, h[2][2], table ());
      generate (Define (fun, Begin (Return (ret))));
    }
    else {
      string err ("\"Fatal error: Generic method " *
		  as_string (S (as_symbol (g[1]))) *
		  " is not supported\\n\"");
      generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
      generic _exit = list (symbol ("exit"), symbol ("1"));
      generate (Define (fun, Begin (_error, _exit)));
    }
  }
}

void
extender::gentype_methods (list l) {
  if (!nil (l)) {
    gentype_methods (car (l));
    gentype_methods (cdr (l));
  }
}

void
extender::produce_gentype () {
  // The gentype is a pointer to gentype_rep
  generate (Type (gentype_rep, MMX_CLASS));
  generate (Typedef (gentype, Pointer (gentype_rep)));

  // Generate the gentype_rep structure
  list l= specialize (imports, geninst (), true);
  l= decl_as_ccpp (l, true);
  generic gt= Type (symbol ("mmxt"), mmx_generic);
  generic gc= Define (Type (gentype_rep, Inline (Constructor (MMX_VOID))),
		      Begin ());
  generic gd= Define (Type (gentype_rep, Inline (Virtual (Destructor ()))),
		      Begin ());
  if (glue_flag) l= cons (gt, gc, gd, l);
  else l= cons (gc, gd, l);
  generate (make_struct (gentype_rep, l));

  // Constructing gentypes from Mathemagix types
  if (glue_flag) {
    generic x   = symbol ("g");
    generic ret = symbol ("ret");
    generic fnm = symbol ("mmx_type_to_" * the_prefix * "_type");
    generic ft  = Function (Cross (Type (x, mmx_generic)), gentype);
    generic ptr = New (gentype_rep, list ());
    generic defp= Assign (Type (ret, Pointer (gentype_rep)), ptr);
    generic sett= Assign (Arrow (ret, symbol ("mmxt")), x);
    generic body= Begin (defp, sett, Return (ret));
    generate (Define (Type (fnm, ft), body));

    generic as  = symbol ("as_" * the_prefix * "_type");
    generic asgl= mmx_symbol ("as_glue_type");
    generic vw  = mmx_specialize ("view_as", gentype);
    generic setr= Assign (Type (ret, mmx_generic), list (asgl, x));
    generic ifst= If (list (mmx_symbol ("nil"), ret),
		      Return (list (fnm, x)),
		      Return (list (vw, ret)));
    generate (Define (Type (as, ft), Begin (setr, ifst)));
  }

  // Printing geninst instances should never happen,
  // but the routine should be included to make the compiler happy
  if (glue_flag) {
    generic x  = symbol ("g");
    generic out= symbol ("out");
    generic ost= Alias (MMX_OSTREAM);
    generic dom= Cross (Type (out, ost), Type (x, geninst ()));
    generic fun= Type (MMX_LESSLESS, Function (dom, ost));
    string err ("\"Fatal error: << should not be called for generics\"");
    generic _error= list (MMX_LESSLESS, symbol ("cerr"), symbol (err));
    generic _exit = list (symbol ("exit"), symbol ("1"));
    generate (Define (fun, Begin (_error, _exit)));
  }

  // Generate the methods
  l= specialize (imports, geninst (), false);
  gentype_methods (l);
}

/******************************************************************************
* Produce the generic instance types, indexed by numbers 1, 2, ....
* These types are instances of dynamic types. curinst(i) holds
* the current type of geninst(i).
******************************************************************************/

static generic
converter_class (
  generic tmpl_var, generic cvr, generic F, generic T, generic body)
{
  generic x  = symbol ("x");
  generic h  = symbol ("helper");

  // the contents of the converter structure
  generic tx = Type (x, F);
  generic f  = Static (Inline (Function (tx, T)));
  generic b  = Define (Type (h, f), body);
  generic val= Struct (b);

  // the converter class with helper routine
  generic var= Specialize (cvr, T, F);
  generic tp = Forall (Type (tmpl_var, MMX_CLASS), MMX_CLASS);
  if (N(tmpl_var) == 2) {
    if ((F == symbol ("F")) && (T == symbol ("T"))) var= cvr;
    tp = Forall (Type (tmpl_var[0], MMX_CLASS),
		 Forall (Type (tmpl_var[1], MMX_CLASS),
			 MMX_CLASS));
  }
  return Define (Type (var, tp), val);
}

void
extender::produce_geninst () {
  generate (make_blackbox_struct (geninst ()));

  generic x  = symbol ("x");
  generic h  = symbol ("helper");
  generic cv = symbol ("convert");
  generic cvr= symbol ("converter");
  generic Cv = symbol ("cast");
  generic Cvr= symbol ("caster");
  generic F  = symbol ("F");
  generic T  = symbol ("T");

  // The converter meta template
  generic b0= Begin (Return (Transtype (x, T)));
  generate (converter_class (list (T, F), cvr, F, T, b0));

  // From abstract to concrete
  generic a1= Transtype (Dot (x, symbol ("rep")), Pointer (T));
  generic b1= Begin (Return (Unpointer (Brackets (a1))));
  generate (converter_class (T, cvr, geninst (), T, b1));

  // From concrete to abstract
  generic b2= Begin (Return (list (geninst (), New (F, x))));
  generate (converter_class (F, cvr, F, geninst (), b2));

  // From abstract aliases to concrete aliases
  generate (converter_class (T, cvr, Alias (geninst ()), Alias (T), b1));

  // From concrete aliases to abstract aliases
  generic a4= New (geninst (), Unalias (x));
  generic b4= Begin (Return (Unpointer (a4)));
  generate (converter_class (F, cvr, Alias (F), Alias (geninst ()), b4));

  // The main conversion routine
  generic cc= Scope (list (cvr, T, F), h);
  generic ff= Inline (Function (Type (x, F), T));
  generic tt= Forall (Type (T, MMX_CLASS), Forall (Type (F, MMX_CLASS), ff));
  generic bb= Return (list (cc, x));
  generate (Define (Type (cv, tt), Begin (bb)));

  // Crude casting meta template
  generic B0= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, F, T, B0));

  // Crude casting for aliases meta template
  generic B1= Begin (Return (Cast_var (x, T)));
  generate (converter_class (list (T, F), Cvr, Alias (F), Alias (T), B1));

  // The main conversion routine for crude casting
  generic Cc= Scope (list (Cvr, T, F), h);
  generic Bb= Return (list (Cc, x));
  generate (Define (Type (Cv, tt), Begin (Bb)));
}

list
extender::abstract (int nr, list args, list types) {
  // Convert to abstract type when necessary
  if (nil (args)) return args;
  generic x= car (args);
  generic stype= car (types) [2];
  generic dtype= substitute (stype, geninst (nr), geninst ());
  if (dtype != stype)
    x= list (Specialize (symbol ("cast"), dtype, stype), x);
  return cons (x, abstract (nr, cdr (args), cdr (types)));
}

void
extender::geninst_methods (int nr, generic g) {
  if (!is_func (g, MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1]))
    error (g[1], "Expected a symbol");
  else if (!is_func (g[2], MMX_FUNCTION, 2))
    error (g[2], "Expected a function type");
  else if (!is_func (g[2][1], MMX_CROSS))
    error (g[2][1], "Expected a cross type");
  else {
    g= name_parameters (g);
    generic name = genmethod (g[1]);
    generic fun  = Arrow (curtype (1), name);
    list    args = arg_list (1, N(g[2][1]));
    generic appl = cons (fun, abstract (nr, args, cdr (as_list (g[2][1]))));
    generic dtype= g[2][2];
    generic stype= substitute (dtype, geninst (nr), geninst ());
    if (dtype != stype)
      appl= list (Specialize (symbol ("cast"), dtype, stype), appl);
    if (dtype == MMX_VOID) generate (Define (g, Begin (appl)));
    else generate (Define (g, Begin (Return (appl))));
  }
}

void
extender::geninst_methods (int nr, list l) {
  if (!nil (l)) {
    geninst_methods (nr, car (l));
    geninst_methods (nr, cdr (l));
  }
}

void
extender::produce_geninst (int nr) {
  generate (Type (curtype (nr), Static (gentype)));
  generate (make_blackbox_struct (geninst (nr)));
  list l= specialize (imports, geninst (nr), false);
  l= decl_as_ccpp (l);
  geninst_methods (nr, l);
}

} // namespace mmx
